
/*	Copyright (C) G. Jespersen
	See Licence.txt.*/

#include "Stdafx.h"
#include "Khan.h"
#include "JsonParse.h"


#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif


// PROFILING
//
//	To run file object only - use -
//	/EXCALL /INC JsonParse.obj
//
//	To run ParseLibrary() only - use -
//	/SF ?ParseLibrary@CParseJson@@QAEHPAVCTopicTree@@ABVCString@@PAUHWND__@@AAHAAV3@@Z
//	/SF ?TestParseLibrary@CParseJson@@QAEHPAVCTopicTree@@ABVCString@@AAV3@@Z


//////////////////////////////////////////////////////////////////
// CTopicBlk
//	For temo storage of a block of JSON topic data.
//	Gets converted to a Khan tree CTopic.

CTopicBlk::CTopicBlk()
{
	Zero();
}

CTopicBlk::~CTopicBlk()
{
}

CTopicBlk::CTopicBlk(const CTopicBlk & topicBlk)
{
	Copy(topicBlk);
}

void CTopicBlk::operator=(const CTopicBlk & topicBlk)
{
	Copy(topicBlk);
}

void CTopicBlk::Copy(const CTopicBlk & topicBlk)
{
	m_nOpeningBrace  = topicBlk.m_nOpeningBrace;
	m_description	 = topicBlk.m_description;
	m_title			 = topicBlk.m_title;
	m_extendedSlug	 = topicBlk.m_extendedSlug;
	m_id			 = topicBlk.m_id;
	m_kind			 = topicBlk.m_kind;
}

void CTopicBlk::Zero()
{
	m_nOpeningBrace = -1;
	m_description = _T("");
	m_title.Empty();
	m_extendedSlug = _T("");
	m_id = _T("");
	m_kind = _T("");
}

BOOL CTopicBlk::ToTreeTopic(CTopic & khanTopic, CStringArray & nodesAr)
{
	// Must have.
	if(m_extendedSlug.IsEmpty())
		return FALSE;

	CString str = m_extendedSlug;

	// Ensure lower case.
	str.MakeLower();

	// Parse slug: e.g "math/algebra/absolute-value"
	CString sub;
	int i;
	while(( i = str.Find(_T("/"), 0) ) != -1)
	{
		sub = str.Left(i);
		str.Delete(0, i + 1);
		nodesAr.Add(sub);
	}
	// Add the last
	nodesAr.Add(str);

	// Get num topic nodes.
	int nNodes = nodesAr.GetSize();
	if(!nNodes)
		return FALSE;

	// Must have a title.
	if(m_title.IsEmpty())
		return FALSE;
   
	// Set the topic's title and description.
	khanTopic.title = m_title;
	khanTopic.description = m_description;

	// Set this topics tree topic to the last id array element.
	khanTopic.treeTopic = nodesAr[nNodes - 1];

	// Set this topics parent tree topic.
	if(nNodes > 1)
		khanTopic.parentTreeTopic = nodesAr[nNodes - 2];

	// Done. Caller to add videos and add topic to tree.
	return TRUE;
}

///////////////////////////////////////////////////////////////////
// CParseJson

CParseJson::CParseJson()
{
}

CParseJson::~CParseJson()
{
}

// Parse the json data saved on disk at pathIn param.
//
// Relies on properly formatted JSON (Just-Stupid-Object-Notation).
//
// For use with khan api calls:
//	http://www.khanacademy.org/api/v1/topictree
//
// Use "Test" versions for debugging. See CDataView::OnTest().
// See "Test" version for comments and changes.
//
// Revised 05/26/2017
// Revised 10/04/2017
// 
BOOL CParseJson::ParseLibrary(CTopicTree * tree, const CString & pathIn,
							  HWND hWndDlg, int & progPos, CString & errMsg)
{
	// File for reading.
	CFile datafile;
	CFileException e;
	if(!datafile.Open(pathIn, CFile::modeRead | CFile::typeBinary, &e))
	{
		TCHAR szError[1024];
		e.GetErrorMessage(szError, 1024);
		CString er;
		er.Format(_T("Couldn't open file: %s"), szError);
		TRACE(er);
		errMsg = er;
		return FALSE;
	}

	TCHAR	ch			= _T('0');
	TCHAR	prev_ch		= _T('0');
	TCHAR	prevPrev_ch = _T('0');
	int		iCh			= 0;
	UINT	nRead		= 0;
	BOOL	inString	= FALSE;
	BOOL	getKeyword	= TRUE;
	int		nBrackets	= 0;			// not used

	m_skipBlock				= FALSE;	// Run through some blocks - 3815128 chars skipped
	m_skipBlockStartBrace	= 0;
	m_braceOffset			= -1;		// Brace count offset to topic blocks
	m_lastBraceCloseFP		= 0;		// File pointer distance between } and {
	m_lastBraceOpenFP		= 0;		// File pointer distance between { and }
	m_numBraces				= 0;

	// Error stats.
	// Used in global KhanUpdateThread() in Doc.cpp
	// to possibly abort update.
	m_nBraceMismatch = 0;
	m_nNodeBraceError = 0;

	// Unicode 10/04/2017 search
	const int uniBufSz = 5;
	BYTE uniBuf[uniBufSz];
	int nUniChars = 0;
	int nSeqLen = 0;
	BOOL uniNext = FALSE;
	BOOL uniTrash = FALSE;
	ZeroMemory(uniBuf, uniBufSz);

	// For non-string null value test.
	m_nCharsBetween = 0;
	CString strBetween;

	// Buffer.
	const UINT bufSz = 4096;
	BYTE buf[bufSz];				// 10/04/2017 search

	// Running file pos.
	m_filePos = 0;

	// Read chunks.
	while((nRead = datafile.Read(buf, bufSz)) > 0)
	{
		// Parse the chunk.
		for(UINT i = 0; i < nRead; i++)
		{
			m_filePos ++;
			ch = buf[i];
			iCh = buf[i];

			// Run through unneeded blocks.
			// Revised 10/04/2017 search
			if(m_skipBlock)
			{
				// Watch for empty arrays. 
				if(m_skipBlockStartBrace == -1 &&
				   ( iCh == 91 || iCh == 93 )  )	// []
				{
					m_skipBlock = FALSE;
					continue;
				}
				// Start of skipped block.
				if(iCh == 123)	// {
				{
					m_numBraces ++;
					if(m_skipBlockStartBrace == -1)
						m_skipBlockStartBrace = m_numBraces;
				}
				// End of skipped block.
				else if(iCh == 125)	// }
				{
					m_numBraces --;
					if(m_numBraces < m_skipBlockStartBrace)
					{
						m_skipBlock = FALSE;
						getKeyword = TRUE;
					}
				}
				continue;
			}

			// Unicode chars. 10/04/2017 search
			// See TestParseLibrary() version for notes.
			if(iCh < 0 || iCh > 127)
			{
				// Flag
				uniTrash = FALSE;

				// Is the next byte a uni byte?
				uniNext = FALSE;
				if(i + 1 < nRead && buf[i + 1] > 127)
					uniNext = TRUE;

				// Get sequence length and first byte.
				if(nUniChars == 0)
				{
					if( (iCh & 0x80) == 0 )			// first 128 bytes are ascii -
						nSeqLen = 1;				//  should not get here
					else if( (iCh & 0xE0) == 0xC0 )
						nSeqLen = 2;
					else if( (iCh & 0xF0) == 0xE0 )
						nSeqLen = 3;
					else if( (iCh & 0xF8) == 0xF0 )
						nSeqLen = 4;
					else
						nSeqLen = 0;
					if(nSeqLen && uniNext)
					{
						uniBuf[nUniChars] = (BYTE)ch;
						nUniChars ++;
						continue;
					}
				}
				// Get next byte.
				else if(nUniChars && nUniChars < nSeqLen)
				{
					uniBuf[nUniChars] = (BYTE)ch;
					nUniChars ++;

					if(nUniChars < nSeqLen)
						continue;

					else if(nUniChars == nSeqLen)
					{
						// Get as a single wchar_t (unsigned short).
						TCHAR t;
						int ret = MultiByteToWideChar(CP_UTF8,
													  0,
													  (LPCSTR)uniBuf, nUniChars, 
													  &t, 1);
						// Take the uni char.
						if(ret)
						{
							// Bad chars....
							//
							// 8203 ZERO WIDTH SPACE - needed?
							if(t == 8203)
								uniTrash = TRUE;
							// Swap 160 Non-breaking space for normal space 32.
							else if(t == 160)
								ch = _T(' ');
							// Else set TCHAR for use below.
							else
								ch = t;
						}
						// No conversion, discard the char.
						else
						{
							// Get next. Note there is about 28253 sequences processed,
							// and about 1161 errors here. These are probably unicode 
							// surrogates/supplementary characters that we will not 
							// deal with for now.
							uniTrash = TRUE;
						}
					}
				}
				// Clean for next
				if(nUniChars)
				{
					nUniChars = 0;
					nSeqLen = 0;
					ZeroMemory(uniBuf, uniBufSz);
					// Get next char if trash.
					if(uniTrash)
						continue;
				}
			}

			// Switch the char.
			switch(iCh)
			{
				// Brackets - not used.
				case 91:	// [
				{
					nBrackets ++;
					break;
				}

				case 93:	// ]
				{
					nBrackets --;
					break;
				}

				// Braces.
				case 123:	// {
				{
					// JSON block begining.
					// Note m_numBraces is incremented in BeginBlock
					if(!inString)
					{
						// Start a new Block.
						BlockBegin();
					}

					// Else, if the { is part of a m_keyval.value string.
					else if(inString && !getKeyword)
					{
						m_keyval.value += ch;

						// Get next char.
						prevPrev_ch = prev_ch;
						prev_ch = ch;
						break;
					}
				}
				break;

				case 125:	// }
				{
					// JSON block ending.
					// Note m_numBraces decremented in EndBlock().
					if(!inString)
					{
						BlockEnd(tree, hWndDlg, progPos);
					}

					// Else, if the } is part of a m_keyval.value string.
					else if(inString && !getKeyword)
					{
						m_keyval.value += ch;

						// Get next char.
						prevPrev_ch = prev_ch;
						prev_ch = ch;
						break;
					}
				}
				break;				
				
				// Escape sequence next.
				case 92:	// "\" 
				{
					// Save previous two.
					prevPrev_ch = prev_ch;
					prev_ch = ch;				
				}
				break;

				// Start or end of a string.
				case 34:	// "
				{
					// String is complete.
					if(inString)
					{
						// If an escaped single quote ", 
						if(prev_ch == _T('\\'))
						{
							// Get a ref.
							CString & str = getKeyword ? m_keyval.keyword : m_keyval.value;

							// If there is an escaped escape, add a "\"
							// and save. Example: "key": "\\". v0.9.2.0
							if(prevPrev_ch == _T('\\'))
							{
								str += _T('\\');
								inString = FALSE;
							}

							// Else, the " is part of a string.
							else
							{
								str += ch;

								// Get next char.
								prevPrev_ch = prev_ch;
								prev_ch = ch;
								break;
							}

							// Prep for init below.
							prev_ch = _T('0');
						}

						// If getting a keyword.
						if(getKeyword)
						{
							// If valid keyword, get a value next.
							if(IsKeyword())
							{
								getKeyword = FALSE;

								// Begin saving chars after a keyword.
								m_nCharsBetween = 0;
								strBetween = _T("");
							}
							// Else get another keyword.
							else
							{
								m_keyval.Zero();
							}
							// String is cessed.
							inString = FALSE;
						}

						// Getting a value.
						else
						{
							// Process the KEYVAL pair.
							ProcessKeyVal(tree);

							// String processed. Prepare for next.
							m_keyval.Zero();
							getKeyword = TRUE;
							inString = FALSE;
						}
					}

					// Else begin a string.
					else
					{
						// If begining to get a Key Value, check for a 
						// non-string null key value.
						// Note the distance between a keyWord and a keyValue
						// is usually 2 chars.
						if(!getKeyword && m_nCharsBetween > 4) // ": null,"
						{
							// If a null key value, we are getting a new keyWord.
							if( ( strBetween.Find(_T("null"), 0) ) != -1 )
							{
								getKeyword = TRUE;
								m_nCharsBetween = 0;
								strBetween = _T("");
							}
						}
						// Begining a string.
						inString = TRUE;
					}

					// Save previous chars.
					prevPrev_ch = prev_ch;
					prev_ch = ch;				
				}
				break;

				// Get chars.
				default:
				{
					// Get string data.
					if(inString)
					{
						// Flag for bad escape sequence.
						BOOL addChar = TRUE;

						// Escape sequences as bytes. See case 92 above. 10/04/2017 search.
						if(prev_ch == 92) // '\'
						{
							// Newline (line feed) 'n'. Add a newline escape.
							if(ch == 110)
								ch = 10;

							// Carriage Return 'r'. Should be followed by a \n.
							else if(ch == 114)
								addChar = FALSE;

							// Form Feed - f
							else if(ch == 102)
								addChar = FALSE;

							// Foreward slash
							else if(ch == 47)
								ch = 47;

							// $ sign.
							else if(ch == 36)
								ch = 36;

							// Unhandled  /* \t \s */
							// Add the slash, and add the char below.
							else
							{
								if(getKeyword)
									m_keyval.keyword += _T('\\');
								else
									m_keyval.value += _T('\\');
							}
						}
						// Add the char or the escape sequence char.
						if(addChar)
						{
							if(getKeyword)
								m_keyval.keyword += ch;
							else
								m_keyval.value += ch;
						}
					}
					// Else get a video's "date_added" - a non-string value.
					else if(m_keyval.valueIsNumeric)
					{
						// Should end with a comma.
						if(ch == _T(','))
						{
							// Save the value.
							ProcessKeyVal(tree);

							// Prepare for next.
							m_keyval.Zero();
							getKeyword = TRUE;
						}
						// Save up the non-string chars.
						else
							m_keyval.value += ch;
					}
					// Else gather up non string chars following a keyword.
					else if(!getKeyword)
					{
						m_nCharsBetween ++;
						strBetween += ch;
					}

					// Save char as previous
					prevPrev_ch = prev_ch;
					prev_ch = ch;					
				}
				break;				
			}//switch
		}
	}
	return TRUE;
}

// A { has been encountered. 
// Begin a JSON Block. May be a Topic or Video.
BOOL CParseJson::BlockBegin()
{
	// For each brace, we add a block to local array.
	// Note it's possible this block will only get 
	// its m_nOpeningBrace initialized.

	// First, check if there is any local data available 
	// before this {, for use with the previous block.
	int sz = m_topicBlks.GetSize();
	if(sz && m_topicBlks[sz - 1].m_nOpeningBrace == m_numBraces)
	{
		CTopicBlk & blk = m_topicBlks[sz -1];
		TestGetTopicBlockData(blk);

		// Since the local vars have now been stored in
		// the previous block, zero them or they can get read later
		// in the next child block - for example a "curation"
		// block inside a topic.
		ZeroParseVars();
	}

	// Increment for new brace.
	m_numBraces ++;

	// Init a new empty blk for this brace
	// with the brace member set.
	m_tempBlk.Zero();
	m_tempBlk.m_nOpeningBrace = m_numBraces;

	// Add new blk to local array.
	m_topicBlks.Add(m_tempBlk);

	// Save the new opening brace FP.
	m_lastBraceOpenFP = m_filePos;

	// Done
	return TRUE;

}

// A } has been encountered.
BOOL CParseJson::BlockEnd(CTopicTree * tree, HWND hWndDlg, int & progPos)
{
	BOOL ret = FALSE;

	// Get the block for this closing brace.
	int sz = m_topicBlks.GetSize();
	if(sz && m_topicBlks[sz - 1].m_nOpeningBrace == m_numBraces)
	{
		CTopicBlk & blk = m_topicBlks[sz -1];

		// Get any saved local block data.
		GetTopicBlockData(blk);

		// If this block is a Video block.
		if(blk.m_kind == _T("Video") &&
		   IsValidCurrentVideo() )
		{
			// Add to array.
			m_videos.Add(m_curVideo);

			// Zero local video vars.
			m_curVideo.Zero();
			m_mp4 = _T("");
			m_duration = _T("");
			m_dateAdded = _T("");
		}

		// If this block is a Topic block.
		else if(blk.m_kind == _T("Topic"))
		{
			// If a valid topic blk.
			if(IsValidTopicBlock(blk))
			{
				CTopic	topic;
				CStringArray nodesAr;
				CString extSlug = blk.m_extendedSlug; // dbg

				// Convert to a CTopic.
				if(blk.ToTreeTopic(topic, nodesAr))
				{
					// 5/26/2017 fix
					// Get the (one time) offset to first topic brace for test below.
					// If the JSON begins with the "khan Academy" root topic.
					// i.e. "slug": "root" and "id": "x00000000", we will have
					// one brace more than nodes.
					if(m_braceOffset == -1)
					{
						int nNodes = nodesAr.GetSize();
						if(m_numBraces && nNodes && (m_numBraces >= nNodes) )
							m_braceOffset = m_numBraces - nNodes;

						// Should not happen.
						else
						{
							m_braceOffset = 0;
							TRACE(_T("ERROR CParseJson::BlockEnd - first node brace count error\n"));
						}
					}

					// Ensure node count matches JSON brace count.
					// Missmatch may be cause to abort.
					int nNodes = nodesAr.GetSize();
					if(nNodes + m_braceOffset == m_numBraces)
					{
						// If videos have been saved, they belong to this topic.
						int nVids = m_videos.GetSize();
						if(nVids)
						{
							for(int i = 0; i < nVids; i++)
							{
								// Videos may not have their parent topic string set.
								m_videos[i].topic = topic.treeTopic;

								// Add to CTopic video array.
								topic.videos.Add(m_videos[i]);
							}
							// Clean the local array.
							m_videos.RemoveAll();
						}

						// Add topic to the tree.
						//
						// If the topic has videos, add to array.
						if(topic.videos.GetSize() > 0)
						{
							int ret = tree->AddTopic(topic, nodesAr);

							// dbg
							if(ret)
							{
								TRACE(_T("CParseJson::BlockEnd - adding topic %s with %d vids.\n"), 
										extSlug, topic.videos.GetSize());
							}
							else
							{
								TRACE(_T("ERROR CParseJson::BlockEnd - AddTopic failed for %s.\n"),
										extSlug);
							}
						}

						// If a valid topic with no videos, it is a parent topic 
						// which needs to be updated. 
						// Note adding child blocks creates a dummy parent from
						// extended slug data.
						else
						{
							// Update Topic Tree Param.
							int ret = tree->UpdateTopic(topic, nodesAr, TRUE);

							// dbg
							if(ret)
							{
								TRACE(_T("CParseJson::BlockEnd - updating topic %s.\n"), 
										extSlug);
							}
							else
							{
								// Updates will fail if topic block only contains Exercises.
								// The extended slug will be unique to the tree so no topic
								// is available to update. About 46 errors on 4/19/15
								TRACE(_T("WARNING CParseJson::BlockEnd - UpdateTopic failed for %s.\n"),
										extSlug);
							}
						}

						// Update dlg progress pos every 20 topics.
						if(hWndDlg)
						{
							progPos ++;
							if(progPos % 20 == 0)
								PostMessage(hWndDlg, WM_USER_PROGRESS_POS, 0, (LPARAM)progPos);
						}
					}
					// Else node position not matching brace position.
					// Possibly cancel update.
					else
					{
						m_nNodeBraceError ++;
						TRACE(_T("ERROR CParseJson::BlockEnd Brace @ slug: %s\n"), 
							extSlug);
					}
				}
				// Could not convert to Tree Topic.
				else
				{
					TRACE(_T("ERROR CParseJson::BlockEnd ToTreeTopic failed\n"));
				}
			}
			// Not a valid topic.
			else
			{
				TRACE(_T("CParseJson::BlockEnd TestIsValidTopicBlock() failed\n"));
			}
		}// topic blk

		// Neither a Video or Topic block. 
		else
		{
			TRACE(_T("CParseJson::BlockEnd useless block\n"));
		}

		// Finish up
	}
	// Block's m_nOpeningBrace != m_numBraces,
	else
	{
		TRACE(_T("ERROR CParseJson::BlockEnd m_nOpeningBrace != m_numBraces\n"));
	}

	// Prepare for next.
	//
	// Clean vars.
	ZeroParseVars();
	// Remove this block from local array (if valid). 10/04/2017 added test
	if(sz)
		m_topicBlks.RemoveAt(sz - 1);
	// Decrement.
	m_numBraces --;

	return ret;
}

// Is the local m_curVideo initialized.
BOOL CParseJson::IsValidCurrentVideo()
{
	if(	m_curVideo.mp4.IsEmpty()			||
		m_curVideo.description.IsEmpty()	||
		m_curVideo.title.IsEmpty()			||
		m_curVideo.date_added.IsEmpty()		||
		m_curVideo.duration.IsEmpty()		)
		 return FALSE;
	// Ok.
	return TRUE;
}

// Set blk param with valid data, if any.
BOOL CParseJson::GetTopicBlockData(CTopicBlk & blk)
{
	// If a local var is valid, save it to the blk
	// and or current video.
	//
	// Kind
	if(!m_kind.IsEmpty())
		blk.m_kind = m_kind;

	// Extended Slug
	if(!m_extendedSlug.IsEmpty())
		blk.m_extendedSlug = m_extendedSlug;

	// Id
	if(!m_id.IsEmpty())
		blk.m_id = m_id;

	// Description
	if(!m_description.IsEmpty())
	{
		blk.m_description = m_description;
		m_curVideo.description = m_description;
	}

	// Title
	if(! m_title.IsEmpty())
	{
		blk.m_title	= m_title;
		m_curVideo.title = m_title;
	}

	// Date Added
	if( !m_dateAdded.IsEmpty())
		m_curVideo.date_added = m_dateAdded;

	// Mp4
	if( !m_mp4.IsEmpty())
		m_curVideo.mp4 = m_mp4;

	// Duration
	if( !m_duration.IsEmpty())
		m_curVideo.duration = m_duration;

	return TRUE;
}

// Ensure final topic block is valid.
BOOL CParseJson::IsValidTopicBlock(const CTopicBlk & blk)
{
	// Kind is topic.
	if(blk.m_kind != _T("Topic"))
		return FALSE;

	// Must have slug.
	if(blk.m_extendedSlug.IsEmpty())
		return FALSE;

	// Must have an id even though not used
	if(blk.m_id.IsEmpty())
		return FALSE;

	// At least one string
	if(blk.m_description.IsEmpty() &&
	   blk.m_title.IsEmpty()		)
		return FALSE;

	// Not an empty root id.
	if(blk.m_id == _T("x00000000"))
		return FALSE;

	// Ok
	return TRUE;
}

// Check if m_keyval.keyword a Khan Keyword.
BOOL CParseJson::IsKeyword()
{
	// kind - both
	if(m_keyval.keyword == _T("kind"))
		return TRUE;

	// id - both
	else if(m_keyval.keyword == _T("id"))
		return TRUE;	
	
	// title - both 
	else if(m_keyval.keyword == _T("title"))
		return TRUE;

	// description - both
	else if(m_keyval.keyword == _T("description"))
		return TRUE;

	// date_added - video
	else if(m_keyval.keyword == _T("date_added"))
		return TRUE;

	// mp4 - video
	else if(m_keyval.keyword == _T("mp4"))
		return TRUE;

	// duration - video
	else if(m_keyval.keyword == _T("duration"))
	{
		m_keyval.valueIsNumeric = TRUE;
		return TRUE;
	}

	// Skip block - short blocks easily skipped 
	// Revised 10/04/2017 search
	else if(m_keyval.keyword == _T("thumbnail_urls")	||
			m_keyval.keyword == _T("concept_tags_info") || //no has arrays
			m_keyval.keyword == _T("curation")			)
	{
		// If not already skipping.
		if(!m_skipBlock)
		{
			m_skipBlock = TRUE;
			m_skipBlockStartBrace = -1;
		}
		// Return FALSE to get another keyword.
	}

	// extended_slug - topic
	else if(m_keyval.keyword == _T("extended_slug"))
		return TRUE;

	return FALSE;
}

// Process a KEYVAL pair.
BOOL CParseJson::ProcessKeyVal(CTopicTree * tree)
{
	// kind - both
	if(m_keyval.keyword == _T("kind"))
	{
		m_kind = m_keyval.value;
	}

	// id - both
	else if(m_keyval.keyword == _T("id"))
	{
		m_id = m_keyval.value;
	}

	// title - both  
	else if(m_keyval.keyword == _T("title"))
	{
		m_title = m_keyval.value;
		m_title.TrimRight();
	}

	// description - both
	else if(m_keyval.keyword == _T("description"))
		m_description = m_keyval.value;

	// date_added - video
	else if(m_keyval.keyword == _T("date_added"))
		m_dateAdded = m_keyval.value;

	// mp4 - video
	else if(m_keyval.keyword == _T("mp4"))
		m_mp4 = m_keyval.value;

	// duration - video
	else if(m_keyval.keyword == _T("duration"))
	{
		m_duration = m_keyval.value;
		// Clean up string, e.g., ": 339"
		m_duration.Remove(_T(':'));
		m_duration.TrimLeft();
		m_duration.TrimRight();
	}

	// extended_slug - topic
	else if(m_keyval.keyword == _T("extended_slug"))
	{
		 m_extendedSlug = m_keyval.value;
	}

	return TRUE;
}

// Prepare Current Block data.
void CParseJson::ZeroParseVars()
{
	// Local vars used for parsing.
	m_extendedSlug	= _T("");	// topic

	m_dateAdded		= _T("");	// video
	m_mp4			= _T(""); 	// video
	m_duration		= _T(""); 	// video

	m_description	= _T("");	// both
	m_title			= _T("");	// both
	m_id			= _T("");	// both
	m_kind			= _T("");	// both

	//m_curVideo.Zero();		// 6/2/16 no - used to store data from URL block

	// Zero'd while parsing:
	//
	// m_inVideoUrlBlock;	
	// m_keyval;
	// m_numBraces;
	// m_nCharsBetween;
}


/////////////////////////////////////////////////////////////////////////
// Test versions - called from CView::Test for debugging.

// 10/28/17 4544.435 millisecond with skipblock - release khan.dat 76.9 MB
//			5917.597 millisecond without skipblock
//

/* 6/2/16 profine release - khan.dat 46.6 MB
    Time in module: 3683.651 millisecond
        Func          Func+Child           Hit
        Time   %         Time      %      Count  Function
---------------------------------------------------------
     461.059  11.7     3927.387 100.0        1 CParseJson::TestParseLibrary(class CTopicTree *,class CString const &,class CString &) (jsonparse.obj)
    2789.329  71.0     2789.329  71.0 38915582 CString::operator+=(unsigned short) (mfc42u.dll)
     280.222   7.1      280.222   7.1  4123781 CString::operator=(unsigned short const *) (mfc42u.dll)
       6.895   0.2      155.988   4.0    64517 CParseJson::TestBlockEnd(class CTopicTree *,struct HWND__ *,int &) (jsonparse.obj)
       4.410   0.1      181.069   4.6    64517 CParseJson::TestBlockBegin(void) (jsonparse.obj)
     103.748   2.6      103.748   2.6  1174424 CParseJson::TestIsKeyword(void) (jsonparse.obj)
      11.865   0.3       80.699   2.1   129033 CParseJson::TestZeroParseVars(void) (jsonparse.obj)
       3.225   0.1       88.952   2.3    64517 CArray<class CTopicBlk,class CTopicBlk>::Add(class CTopicBlk) (jsonparse.obj)
      52.210   1.3       52.210   1.3  1737063 CString::operator=(class CString const &) (mfc42u.dll)
       8.521   0.2       40.768   1.0   129034 CTopicBlk::CTopicBlk(class CTopicBlk const &) (jsonparse.obj)
       7.435   0.2       43.441   1.1   129034 CTopicBlk::Zero(void) (jsonparse.obj)
       4.236   0.1       56.554   1.4    64517 CArray<class CTopicBlk,class CTopicBlk>::SetAtGrow(int,class CTopicBlk) (jsonparse.obj)
*/
BOOL CParseJson::TestParseLibrary(CTopicTree * tree, const CString & pathIn, CString & errMsg)
{
	// Params used by non test version.
	HWND hWndDlg = NULL;
	int progPos = 0;

	// File for reading.
	CFile datafile;
	CFileException e;
	if(!datafile.Open(pathIn, CFile::modeRead | CFile::typeBinary, &e))
	{
		TCHAR szError[1024];
		e.GetErrorMessage(szError, 1024);
		CString er;
		er.Format(_T("Couldn't open file: %s"), szError);
		TRACE(er);
		errMsg = er;
		return FALSE;
	}

	TCHAR	ch			= _T('0');		// unsigned short, 0 to 65,535
	TCHAR	prev_ch		= _T('0');
	TCHAR	prevPrev_ch = _T('0');
	int		iCh			= 0;
	UINT	nRead		= 0;
	BOOL	inString	= FALSE;
	BOOL	getKeyword	= TRUE;
	int		nBrackets	= 0;			// not used

	// 6/2/16
	m_skipBlock = FALSE;
	m_skipBlockStartBrace = 0;
	UINT charsSkipped = 0; //dbg

	m_braceOffset			= -1;		// Brace count offset to topic blocks
	m_lastBraceCloseFP		= 0;		// File pointer distance between } and {
	m_lastBraceOpenFP		= 0;		// File pointer distance between { and }

	// Brace count.
	m_numBraces			= 0;

	// Error stats.
	// Used in global KhanUpdateThread() in Doc.cpp
	// to possibly abort update.
	m_nBraceMismatch = 0;
	m_nNodeBraceError = 0;

	// Unicode 10/04/2017 search
	const int uniBufSz = 5;
	BYTE uniBuf[uniBufSz];
	int nUniChars = 0;
	int nSeqLen = 0;
	BOOL uniNext = FALSE;
	BOOL uniTrash = FALSE;
	ZeroMemory(uniBuf, uniBufSz);
	int nUniOk = 0, nUniErr = 0, nUniExt = 0; // dbg
	//GetCPInfoEx

	// For non-string null value test.
	m_nCharsBetween = 0;
	CString strBetween;

	// Buffer.
	const UINT bufSz = 4096;
	BYTE buf[bufSz];					// BYTE, unsigned char, 0 to 255 10/04/2017 search
	UINT nBuf = 0;//dbg

	// Running file pos.
	m_filePos = 0;

	// Read chunks.
	while((nRead = datafile.Read(buf, bufSz)) > 0)
	{
		// Parse the chunk.
		for(UINT i = 0; i < nRead; i++)
		{
			m_filePos ++;
			ch = buf[i];
			iCh = buf[i];

			// Run through unneeded blocks.
			// See TestIsKeyword().
			// Revised 10/04/2017 search
			if(m_skipBlock)
			{
				// Watch for empty arrays. 
				// For example - concept_tags_info":[] has no curly braces.
				if(m_skipBlockStartBrace == -1 &&
				   ( iCh == 91 || iCh == 93 )  )	// []
				{
					m_skipBlock = FALSE;
					continue;
				}
				// Start of skipped block.
				if(iCh == 123)	// {
				{
					m_numBraces ++;
					if(m_skipBlockStartBrace == -1)
						m_skipBlockStartBrace = m_numBraces;
				}
				// End of skipped block.
				else if(iCh == 125)	// }
				{
					m_numBraces --;
					if(m_numBraces < m_skipBlockStartBrace)
					{
						m_skipBlock = FALSE;
						getKeyword = TRUE;	// already TRUE from case 34
					}
				}
				charsSkipped ++; // dbg 6720822 total
				continue;
			}

			// Unicode chars. 10/04/2017 search
			if(iCh < 0 || iCh > 127)
			{
				// Flag
				uniTrash = FALSE;

				// Is the next byte a uni byte?
				uniNext = FALSE;
				if(i + 1 < nRead && buf[i + 1] > 127)
					uniNext = TRUE;

				// Get sequence length and first byte.
				if(nUniChars == 0)
				{
					if( (iCh & 0x80) == 0 )			// first 128 bytes are ascii -
						nSeqLen = 1;				//  should not get here
					else if( (iCh & 0xE0) == 0xC0 )
						nSeqLen = 2;
					else if( (iCh & 0xF0) == 0xE0 )
						nSeqLen = 3;
					else if( (iCh & 0xF8) == 0xF0 )
						nSeqLen = 4;
					else
						nSeqLen = 0;
					if(nSeqLen && uniNext)
					{
						uniBuf[nUniChars] = (BYTE)ch;
						nUniChars ++;
						continue;
					}
				}
				// Get next byte.
				else if(nUniChars && nUniChars < nSeqLen)
				{
					uniBuf[nUniChars] = (BYTE)ch;
					nUniChars ++;

					if(nUniChars < nSeqLen)
						continue;

					else if(nUniChars == nSeqLen)
					{
						// Get as a single wchar_t (unsigned short).
						// Note the WC_ERR_INVALID_CHARS flag is only supported 
						// on Windows Vista or later.
						// For CP_UTF8 flag, dwFlags must be zero. 
						TCHAR t;
						int ret = MultiByteToWideChar(CP_UTF8,					// CP_UTF8 CP_ACP
													  0,						// 0, MB_ERR_INVALID_CHARS
													  (LPCSTR)uniBuf, nUniChars, 
													  &t, 1);
						// Take the uni char.
						if(ret)
						{
							// Bad chars....
							//
							// 8203 ZERO WIDTH SPACE - needed?
							if(t == 8203)
								uniTrash = TRUE;

							// Swap 160 Non-breaking space for normal space 32.
							else if(t == 160)
								ch = _T(' ');

							// Printing in dbg as ? - Lucida Console font
							// Dec  Hex		Unicode Name					In Arial Font
							// ----------------------------------------------------------
							// 952			GREEK SMALL LETTER THETA
							// 4113			Myanmar Letter Tha				no
							// 7503	1D4F	MODIFIER LETTER SMALL K			no
							// 7504	1D50	MODIFIER LETTER SMALL M			no
							// 8203			ZERO WIDTH SPACE				no
							// 8314			SUPERSCRIPT PLUS SIGN 			no
							// 8315			SUPERSCRIPT HYPHEN-MINUS		no
							// 8531			VULGAR FRACTION ONE THIRD		
							// 8532			Vulgar Fraction Two Thirds	
							// 8536			VULGAR FRACTION FOUR FIFTHS	
							// 8539			Vulgar Fraction One Eighth	
							// 8800			NOT EQUAL TO					
							// 10005		Multiplication X				no

							// Else set TCHAR for use below.
							else
								ch = t;

							// dbg
							nUniOk ++;
							CString s(t);
					//		if(s != _T(' '))
					//			TRACE(_T("CParseJson::TestParseLibrary %d byte unicode char = %s dec %d\n"), 
					//						nUniChars, s, t);
						}
						// No conversion, discard the char.
						else
						{
							// dbg
							DWORD ret = GetLastError();
							CString errCode;
							if(ret == ERROR_INSUFFICIENT_BUFFER)
								errCode = _T("INSUFFICIENT_BUFFER"); // about 890 - all 4 bytes
							else if(ret == ERROR_INVALID_FLAGS)
								errCode = _T("INVALID_FLAGS");
							else if(ret == ERROR_INVALID_PARAMETER)
								errCode = _T("INVALID_PARAMETER");
							else if(ret == ERROR_NO_UNICODE_TRANSLATION)
								errCode = _T("NO_UNICODE_TRANSLATION");
					//		TRACE(_T("CParseJson::TestParseLibrary unicode convert failed w %d bytes errMsg %s\n"),
					//						nUniChars, errCode);
							nUniErr ++;

							// 244, 128, 128, 128	F4, 80, 80, 80
							// 240, 157, 145, 146	F4, 9D, 91, 92
							// 240, 157, 152, 182	F4, 9D, 98, B6
							// 240, 157, 152, 141	F4, 9D, 98, 8D
							//

							// Get next. Note there is about 28253 sequences processed,
							// and about 1161 errors here. These are probably unicode 
							// surrogates/supplementary characters that we will not 
							// deal with for now.
							uniTrash = TRUE;
						}
					}
				}
				// Clean for next
				if(nUniChars)
				{
					nUniChars = 0;
					nSeqLen = 0;
					ZeroMemory(uniBuf, uniBufSz);
					// Get next char if trash.
					if(uniTrash)
						continue;
				}
			}

			// Switch the char.
			switch(iCh)
			{
				// Brackets - not used.
				case 91:	// [
				{
					nBrackets ++;
					break;
				}

				case 93:	// ]
				{
					nBrackets --;
					break;
				}

				// Braces.
				case 123:	// {
				{
					// 6/2/16
					// out 10/04/2017 search
				/*	if(m_skipBlockStartBrace == m_filePos)
					{
						m_skipBlockStartBrace = 0;
						m_skipBlock = TRUE;
						continue;
					}*/

					// JSON block begining.
					// Note m_numBraces is incremented in BeginBlock
					if(!inString)
					{
						// Start a new Block.
						TestBlockBegin();
					}

					// Else, if the { is part of a m_keyval.value string.
					else if(inString && !getKeyword)
					{
						m_keyval.value += ch;

						// Get next char.
						prevPrev_ch = prev_ch;
						prev_ch = ch;
						break;
					}
				}
				break;

				case 125:	// }
				{
					// JSON block ending.
					// Note m_numBraces decremented in EndBlock().
					if(!inString)
					{
						TestBlockEnd(tree, hWndDlg, progPos);
					}

					// Else, if the } is part of a m_keyval.value string.
					else if(inString && !getKeyword)
					{
						m_keyval.value += ch;

						// Get next char.
						prevPrev_ch = prev_ch;
						prev_ch = ch;
						break;
					}
				}
				break;				
				
				// Escape sequence next.
				case 92:	// "\" 
				{
					// Save previous two.
					prevPrev_ch = prev_ch;
					prev_ch = ch;				
				}
				break;

				// Start or end of a string.
				case 34:	// "
				{
					// String is complete.
					if(inString)
					{
						// If an escaped single quote ", 
						if(prev_ch == _T('\\'))
						{
							// Get a ref.
							CString & str = getKeyword ? m_keyval.keyword : m_keyval.value;

							// If there is an escaped escape, add a "\"
							// and save. Example: "key": "\\". v0.9.2.0
							if(prevPrev_ch == _T('\\'))
							{
								str += _T('\\');
								inString = FALSE;
							}

							// Else, the " is part of a string.
							else
							{
								str += ch;

								// Get next char.
								prevPrev_ch = prev_ch;
								prev_ch = ch;
								break;
							}

							// Prep for init below.
							prev_ch = _T('0');
						}

						// If getting a keyword.
						if(getKeyword)
						{
							// If valid keyword, get a value next.
							if(TestIsKeyword())
							{
								getKeyword = FALSE;

								// Begin saving chars after a keyword.
								m_nCharsBetween = 0;
								strBetween = _T("");
							}
							// Else get another keyword.
							else
							{
								m_keyval.Zero();
							}
							// String is cessed.
							inString = FALSE;
						}

						// Getting a value.
						else
						{
							// Process the KEYVAL pair.
							TestProcessKeyVal(tree);

							// String processed. Prepare for next.
							m_keyval.Zero();
							getKeyword = TRUE;
							inString = FALSE;
						}
					}

					// Else begin a string.
					else
					{
						// If begining to get a Key Value, check for a 
						// non-string null key value.
						// Note the distance between a keyWord and a keyValue
						// is usually 2 chars.
						if(!getKeyword && m_nCharsBetween > 4) // ": null,"
						{
							// If a null key value, we are getting a new keyWord.
							if( ( strBetween.Find(_T("null"), 0) ) != -1 )
							{
								getKeyword = TRUE;
								m_nCharsBetween = 0;
								strBetween = _T("");
							}
						}

						// Begining a string.
						inString = TRUE;
					}

					// Save previous chars.
					prevPrev_ch = prev_ch;
					prev_ch = ch;				
				}
				break;

				// Get chars.
				default:
				{
					// Get string data.
					if(inString)
					{
						// Flag for bad escape sequence.
						BOOL addChar = TRUE;

						// Escape sequences as bytes. See case 92 above. 10/04/2017 search.
						if(prev_ch == 92) // '\'
						{
							// Newline (line feed) 'n'. Add a newline escape.
							if(ch == 110)
								ch = 10;

							// Carriage Return 'r'. Should be followed by a \n.
							else if(ch == 114)
								addChar = FALSE;

							// Form Feed - f
							else if(ch == 102)
								addChar = FALSE;

							// Foreward slash
							else if(ch == 47)
								ch = 47;

							// $ sign.
							else if(ch == 36)
								ch = 36;

							// Unhandled  /* \t \s */
							// Add the slash, and add the char below.
							else
							{
								if(getKeyword)
									m_keyval.keyword += _T('\\');
								else
									m_keyval.value += _T('\\');
							}
						}
						// Add the char or the escape sequence char.
						if(addChar)
						{
							if(getKeyword)
								m_keyval.keyword += ch;
							else
								m_keyval.value += ch;
						}
					}
					// Else get a video's "date_added" - a non-string value.
					else if(m_keyval.valueIsNumeric)
					{
						// Should end with a comma.
						if(ch == _T(','))
						{
							// Save the value.
							TestProcessKeyVal(tree);

							// Prepare for next.
							m_keyval.Zero();
							getKeyword = TRUE;
						}
						// Save up the non-string chars.
						else
							m_keyval.value += ch;
					}
					// Else gather up non string chars following a keyword.
					else if(!getKeyword)
					{
						m_nCharsBetween ++;
						strBetween += ch;
					}

					// Save char as previous
					prevPrev_ch = prev_ch;
					prev_ch = ch;					
				}
				break;				
			}//switch
		}
	}
	// dbg
	// nUniOk = 28253, nUniErr = 1161, nUniExt = 0
	TRACE(_T("CParseJson::ParseLibrary unicode stats: Ok=%d, Err=%d, extended chars=%d\n"),
							nUniOk, nUniErr, nUniExt); 
	return TRUE;
}

// A { has been encountered. 
// Begin a JSON Block. May be a Topic or Video.
BOOL CParseJson::TestBlockBegin()
{
	// For each brace, we add a block to local array.
	// Note it's possible this block will only get 
	// its m_nOpeningBrace initialized.

	// First, check if there is any local data available 
	// before this {, for use with the previous block.
	int sz = m_topicBlks.GetSize();
	if(sz && m_topicBlks[sz - 1].m_nOpeningBrace == m_numBraces)
	{
		CTopicBlk & blk = m_topicBlks[sz -1];
		TestGetTopicBlockData(blk);

		// Since the local vars have now been stored in
		// the previous block, zero them or they can get read later
		// in the next child block - for example a "curation"
		// block inside a topic.
		TestZeroParseVars();
	}

	// Increment for new brace.
	m_numBraces ++;

	// Init a new empty blk for this brace
	// with the brace member set.
	m_tempBlk.Zero();
	m_tempBlk.m_nOpeningBrace = m_numBraces;

	// Add new blk to local array.
	m_topicBlks.Add(m_tempBlk);

	// Save the new opening brace FP.
	m_lastBraceOpenFP = m_filePos;

	// Done
	return TRUE;
}

// A } has been encountered.
// return value not used.
BOOL CParseJson::TestBlockEnd(CTopicTree * tree, HWND hWndDlg, int & progPos)
{
	BOOL ret = FALSE;

	// Get the block for this closing brace.
	int sz = m_topicBlks.GetSize();
	if(sz && m_topicBlks[sz - 1].m_nOpeningBrace == m_numBraces)
	{
		CTopicBlk & blk = m_topicBlks[sz -1];

		// Get any saved local block data.
		TestGetTopicBlockData(blk);

		// If this block is a Video block.
		if(blk.m_kind == _T("Video") &&
		   TestIsValidCurrentVideo() )
		{
			// Add to array.
			m_videos.Add(m_curVideo);

			// Zero local video vars.
			m_curVideo.Zero();
			m_mp4 = _T("");
			m_duration = _T("");
			m_dateAdded = _T("");
		}

		// If this block is a Topic block.
		else if(blk.m_kind == _T("Topic"))
		{
			// If a valid topic blk.
			if(TestIsValidTopicBlock(blk))
			{
				CTopic	topic;
				CStringArray nodesAr;
				CString extSlug = blk.m_extendedSlug; // dbg

				// Convert to a CTopic.
				if(blk.ToTreeTopic(topic, nodesAr))
				{
					// 5/26/2017 fix
					// Get the (one time) offset to first topic brace for test below.
					// If the JSON begins with the "khan Academy" root topic.
					// i.e. "slug": "root" and "id": "x00000000", we will have
					// one brace more than nodes.
					if(m_braceOffset == -1)
					{
						int nNodes = nodesAr.GetSize();
						if(m_numBraces && nNodes && (m_numBraces >= nNodes) )
							m_braceOffset = m_numBraces - nNodes;

						// Should not happen.
						else
						{
							m_braceOffset = 0;
							TRACE(_T("ERROR CParseJson::TestBlockEnd - first node brace count error\n"));
						}
					}

					// Ensure node count matches JSON brace count.
					// Missmatch may be cause to abort.
					int nNodes = nodesAr.GetSize();
					if(nNodes + m_braceOffset == m_numBraces)
					{
						// dbg - save raw slugs. See CDataView::OnTest().
						#ifdef _DEBUG
						CString s; 
						for(int iNode = 0; iNode < nNodes; iNode ++) 
						{ 
							s += nodesAr[iNode]; 
							s += _T('/'); 
						}
						m_testSortStrAr.Add(s);
						#endif

						// dbg
						if(nNodes + 1 != m_numBraces)
							int g = 0;

						// If videos have been saved, they belong to this topic.
						int nVids = m_videos.GetSize();
						if(nVids)
						{
							for(int i = 0; i < nVids; i++)
							{
								// Videos may not have their parent topic string set.
								m_videos[i].topic = topic.treeTopic;

								// Add to CTopic video array.
								topic.videos.Add(m_videos[i]);
							}
							// Clean the local array.
							m_videos.RemoveAll();
						}

						// Add topic to the tree.
						//
						// If the topic has videos, add to array.
						if(topic.videos.GetSize() > 0)
						{
							int ret = tree->AddTopic(topic, nodesAr);

							// dbg
							if(ret)
							{
						//		TRACE(_T("CParseJson::TestBlockEnd - adding topic %s with %d vids. FP %u.\n"), 
						//				extSlug, topic.videos.GetSize(), m_filePos);
							}
							else
							{
								TRACE(_T("ERROR CParseJson::TestBlockEnd - AddTopic failed for %s. FP %u.\n"),
										extSlug, m_filePos);
							}
						}

						// If a valid topic with no videos, it is a parent topic 
						// which needs to be updated. 
						// Note adding child blocks creates a dummy parent from
						// extended slug data.
						else
						{
							// Update Topic Tree Param.
							int ret = tree->UpdateTopic(topic, nodesAr, TRUE);

							// dbg
							if(ret)
							{
						//		TRACE(_T("CParseJson::TestBlockEnd - updating topic %s. FP %u.\n"), 
						//				extSlug, m_filePos);
							}
							else
							{
								// Updates will fail if topic block only contains Exercises.
								// The extended slug will be unique to the tree so no topic
								// is available to update. About 46 errors on 4/19/15
							//	TRACE(_T("WARNING CParseJson::TestBlockEnd - UpdateTopic failed for %s FP %u.\n"),
							//			extSlug, m_filePos);
							}
						}

						// Update dlg progress pos every 20 topics.
						if(hWndDlg)
						{
							progPos ++;
							if(progPos % 20 == 0)
								PostMessage(hWndDlg, WM_USER_PROGRESS_POS, 0, (LPARAM)progPos);
						}
					}
					// Else node position not matching brace position.
					// Possibly cancel update.
					else
					{
						m_nNodeBraceError ++;
						TRACE(_T("ERROR CParseJson::TestBlockEnd Brace @ slug: %s FP %u\n"), 
							extSlug, m_filePos);
					}
				}
				// Could not convert to Tree Topic.
				else
				{
					TRACE(_T("ERROR CParseJson::TestBlockEnd ToTreeTopic failed\n"));
				}
			}
			// Not a valid topic.
			else
			{
				//TRACE(_T("CParseJson::TestBlockEnd TestIsValidTopicBlock() failed\n"));
			}
		}// topic blk

		// Neither a Video or Topic block. 
		else
		{
			//TRACE(_T("CParseJson::TestBlockEnd useless block. FP %u\n"), m_filePos);
		}

		// Finish up
	}
	// Block's m_nOpeningBrace != m_numBraces,
	else
	{
		TRACE(_T("ERROR CParseJson::TestBlockEnd m_nOpeningBrace != m_numBraces\n"));
	}

	// CRASH!! 09/07/2017
	// m_numBraces = 0
	// m_topicBlks.m_nSize = 0
	// see test in above block 
	//  - if(sz && m_topicBlks[sz - 1].m_nOpeningBrace == m_numBraces)

	// Prepare for next.
	//
	// Clean vars.
	TestZeroParseVars();
	// Remove this block from local array (if valid). 10/04/2017 added test
	if(sz)
		m_topicBlks.RemoveAt(sz - 1);
	// Decrement.
	m_numBraces --;

	return ret;
}

// new 6/2/16
// Is the local m_curVideo initialized.
BOOL CParseJson::TestIsValidCurrentVideo()
{
	if(	m_curVideo.mp4.IsEmpty()			||
		m_curVideo.description.IsEmpty()	||
		m_curVideo.title.IsEmpty()			||
		m_curVideo.date_added.IsEmpty()		||
		m_curVideo.duration.IsEmpty()		)
		 return FALSE;
	// Ok.
	return TRUE;
}

// Ensure final topic block is valid.
BOOL CParseJson::TestIsValidTopicBlock(const CTopicBlk & blk)
{
	// Kind is topic.
	if(blk.m_kind != _T("Topic"))
		return FALSE;

	// Must have slug.
	if(blk.m_extendedSlug.IsEmpty())
		return FALSE;

	// Must have an id even though not used
	if(blk.m_id.IsEmpty())
		return FALSE;

	// At least one string
	if(blk.m_description.IsEmpty() &&
	   blk.m_title.IsEmpty()		)
		return FALSE;

	// Not an empty root id.
	if(blk.m_id == _T("x00000000"))
		return FALSE;

	// Ok
	return TRUE;
}

// Set blk param with valid data, if any.
// Fail if non topic data present.
BOOL CParseJson::TestGetTopicBlockData(CTopicBlk & blk)
{
	// If a local var is valid, save it to the blk.
	//
	// Kind
	if(!m_kind.IsEmpty())
		blk.m_kind = m_kind;

	// Extended Slug
	if(!m_extendedSlug.IsEmpty())
		blk.m_extendedSlug = m_extendedSlug;

	// Id
	if(!m_id.IsEmpty())
		blk.m_id = m_id;

	// Description
	if(!m_description.IsEmpty())
	{
		blk.m_description = m_description;
		m_curVideo.description = m_description;
	}

	// Title
	if(! m_title.IsEmpty())
	{
		blk.m_title	= m_title;
		m_curVideo.title = m_title;
	}

	// Date Added
	if( !m_dateAdded.IsEmpty())
		m_curVideo.date_added = m_dateAdded;

	// Mp4
	if( !m_mp4.IsEmpty())
		m_curVideo.mp4 = m_mp4;

	// Duration
	if( !m_duration.IsEmpty())
		m_curVideo.duration = m_duration;

	return TRUE;
}

// Check if m_keyval.keyword a Khan Keyword.
BOOL CParseJson::TestIsKeyword()
{
	// kind - both
	if(m_keyval.keyword == _T("kind"))
		return TRUE;

	// id - both
	else if(m_keyval.keyword == _T("id"))
		return TRUE;	
	
	// title - both 
	else if(m_keyval.keyword == _T("title"))
		return TRUE;

	// description - both
	else if(m_keyval.keyword == _T("description"))
		return TRUE;

	// date_added - video
	else if(m_keyval.keyword == _T("date_added"))
		return TRUE;

	// mp4 - video
	else if(m_keyval.keyword == _T("mp4"))
		return TRUE;

	// duration - video
	else if(m_keyval.keyword == _T("duration"))
	{
		m_keyval.valueIsNumeric = TRUE;
		return TRUE;
	}

	// Skip block - short blocks easily skipped 
	// Revised 10/04/2017 search
	else if(m_keyval.keyword == _T("thumbnail_urls")	||
			m_keyval.keyword == _T("concept_tags_info") || //no has arrays
			m_keyval.keyword == _T("curation")			)
	{
		// If not already skipping.
		if(!m_skipBlock)
		{
			m_skipBlock = TRUE;
			m_skipBlockStartBrace = -1;
		}
		// Return FALSE to get another keyword.
	}

	// extended_slug - topic
	else if(m_keyval.keyword == _T("extended_slug"))
		return TRUE;

	return FALSE;
}

// Process a KEYVAL pair.
BOOL CParseJson::TestProcessKeyVal(CTopicTree * tree)
{
	// kind - both
	if(m_keyval.keyword == _T("kind"))
	{
		m_kind = m_keyval.value;

		/* dbg
		if(!m_kind.IsEmpty()			&&
		   ( m_kind != _T("Topic")		&& 
		     m_kind != _T("Video")		&& 
		     m_kind != _T("Separator")	&& 
		     m_kind != _T("Article")	&& 
		     m_kind != _T("Scratchpad") && 
			 m_kind != _T("Exercise") )	 )
		{
			TRACE(_T("Unused Kind %s\n"), m_kind);
		}*/
	}

	// id - both
	else if(m_keyval.keyword == _T("id"))
	{
		m_id = m_keyval.value;
	}

	// title - both  
	else if(m_keyval.keyword == _T("title"))
	{
		m_title = m_keyval.value;
		// Ensure no spaces at end. Fix v0.8.5 - for a few vid postfix errors
		m_title.TrimRight();
	}

	// description - both
	else if(m_keyval.keyword == _T("description"))
		m_description = m_keyval.value;

	// date_added - video
	else if(m_keyval.keyword == _T("date_added"))
		m_dateAdded = m_keyval.value;

	// mp4 - video
	else if(m_keyval.keyword == _T("mp4"))
		m_mp4 = m_keyval.value;

	// duration - video
	else if(m_keyval.keyword == _T("duration"))
	{
		m_duration = m_keyval.value;
		// Clean up string, e.g., ": 339"
		m_duration.Remove(_T(':'));
		m_duration.TrimLeft();
		m_duration.TrimRight();
	}

	// extended_slug - topic
	else if(m_keyval.keyword == _T("extended_slug"))
	{
		 m_extendedSlug = m_keyval.value;
	}

	return TRUE;
}

// Prepare Current Block data.
void CParseJson::TestZeroParseVars()
{
	// Local vars used for parsing.
	m_extendedSlug = _T("");// topic

	m_dateAdded = _T("");	// video
	m_mp4 = _T(""); 		// video
	m_duration = _T(""); 	// video

	m_description = _T("");	// both
	m_title = _T("");		// both
	m_id = _T("");			// both
	m_kind = _T("");		// both

	//m_curVideo.Zero();	// 6/2/16 no - used to store data from URL block

	// Zero'd while parsing:
	//
	// m_inVideoUrlBlock;	
	// m_keyval;
	// m_numBraces;
	// m_nCharsBetween;
}


/////////////////////////////////////////////////////////////////////////
// Utilities
//
// Gross format json data to make it more readable 
// for debugging parse routines.
//
// pathIn is binary, pathOut txt.
void CParseJson::TestUtilityFormat(CString pathIn, CString pathOut)
{
	CString lineStr1, lineStr2, lineStr3;
	
	// File for reading.
	CFile srcfile;
	CFileException e;
	if(!srcfile.Open(pathIn, CFile::modeRead | CFile::typeBinary, &e))
	{
		TCHAR szError[1024];
		e.GetErrorMessage(szError, 1024);
		CString er;
		er.Format(_T("Couldn't open file for reading: %s"), szError);
		TRACE(er);
		return;
	}

	// File for writing.
	CStdioFile destfile;
	if(!destfile.Open(pathOut, CFile::modeWrite | CFile::modeCreate | 
							   CFile::typeText, &e))
	{
		TCHAR szError[1024];
		e.GetErrorMessage(szError, 1024);
		CString er;
		er.Format(_T("Couldn't open file for writing: %s"), szError);
		TRACE(er);
		return;
	}

	const int bufSz = 4096;
	char buf[bufSz];
	TCHAR	c;
	int		nRead;
	CString tab = _T('\t');
	CString crlf = _T('\n');
	int nTabs = 0;
	int t;
	BOOL inString = FALSE;
	int nBrackets = 0;
	CString strNumBracket;
	UINT fp = 0;			// 05/27/16

	// Read chunks.
	while((nRead = srcfile.Read(buf, bufSz)) > 0)
	{
		// Format the chunk.
		for(int i = 0; i < nRead; i++) // 0x0D, 0x0A
		{
			c = buf[i];
			fp++;

			// Inside a string.
			if(c == _T('"'))
				inString = !inString;

			if(c == _T('{') && !inString)
			{
				nBrackets++;
				destfile.WriteString(crlf);

				for(t = 0; t < nTabs; t++)
					destfile.WriteString(tab);

				destfile.Write(&c, 1);

				strNumBracket.Format(_T("-%d"), nBrackets);
				destfile.WriteString(strNumBracket);

				destfile.WriteString(crlf);

				nTabs ++;

				for(t = 0; t < nTabs; t++)
					destfile.WriteString(tab);
			}
			else if(c == _T('}') && !inString)
			{
				nBrackets--;
				nTabs --;

				destfile.WriteString(crlf);

				for(t = 0; t < nTabs; t++)
					destfile.WriteString(tab);

				destfile.Write(&c, 1);

		//		strNumBracket.Format(_T("-%d"), nBrackets);
				strNumBracket.Format(_T("-%d fp %u"), nBrackets, fp);
				destfile.WriteString(strNumBracket);

				destfile.WriteString(crlf);
			}
			else if(c == _T(',') && !inString)
			{
				destfile.Write(&c, 1);
				destfile.WriteString(crlf);
				for(t = 0; t < nTabs; t++)
					destfile.WriteString(tab);
			}

			else
				destfile.Write(&c, 1);
		}
	}
	// Clean up.
	srcfile.Close();
	destfile.Close();
}

// Print slugs	
void CParseJson::TestUtilityPrintSlugs(CString pathOut, BOOL sort)
{	
	// File for writing.
	CStdioFile destfile;
	CFileException e;
	if(!destfile.Open(pathOut, CFile::modeWrite | CFile::modeCreate | 
							   CFile::typeText, &e))
	{
		TCHAR szError[1024];
		e.GetErrorMessage(szError, 1024);
		CString er;
		er.Format(_T("Couldn't open file for writing slugs: %s"), szError);
		TRACE(er);
		return;
	}
	CString crlf = _T('\n');

	int sz = m_testSortStrAr.GetSize();

	// Sort.
	if(sort)
		m_testSortStrAr.SortStrDecend();

	// Loop array.
	for(int i = 0 ; i < sz; i ++)
	{
		destfile.WriteString(m_testSortStrAr[i]);
		destfile.WriteString(crlf);
	}

	// Clean up.
	destfile.Close();
}





