
/*	Copyright (C) G. Jespersen
	See Licence.txt.*/

#include "Stdafx.h"
#include "Khan.h"
#include "KhanJson.h"


#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif


// PROFILING
//
//	To run file object only - use -
//	/EXCALL /INC KhanJson.obj
//
//	To run ParseLibrary() only - use -
//	/SF ?ParseLibrary@CKhanJson@@QAEHABVCString@@@Z


///////////////////////////////////////////////////////////////////
// CKhanJson

CKhanJson::CKhanJson()
{
}

CKhanJson::~CKhanJson()
{
}

// Parse the json data saved on disk at pathIn param.
//
// Relies on properly formatted json (Just-Stupid-Object-Notation).
//
// For use with khan api calls:
//	api/v1/topictree
//
// Use "Test" versions for debugging. See CDataView::OnTest().
//
BOOL CKhanJson::ParseLibrary(CTopicTree * tree, const CString & pathIn,
							 HWND hWndDlg, int & progPos, CString & errMsg)
{
	// File for reading.
	CFile datafile;
	CFileException e;
	if(!datafile.Open(pathIn, CFile::modeRead | CFile::typeBinary, &e))
	{
		TCHAR szError[1024];
		e.GetErrorMessage(szError, 1024);
		CString er;
		er.Format(_T("Couldn't open file: %s"), szError);
		TRACE(er);
		errMsg = er;
		return FALSE;
	}

	TCHAR	ch = _T('0');
	TCHAR	prev_ch = _T('0');
	TCHAR	prevPrev_ch = _T('0');
	int		iCh = 0;
	UINT	nRead = 0;
	BOOL	inString = FALSE;
	BOOL	getKeyword = TRUE;
	int		nBrackets = 0;		// not used

	m_numBraces    = 0;
	m_inVideoBlock = FALSE;
	m_nVidBraces   = 1;
	m_inTopicBlock = FALSE;
	m_nTopicBraces = 1;

	// Error stats v0.9.0
	// Used in global KhanUpdateThread() in Doc.cpp
	// to possibly abort update.
	m_nErrorTopics = 0;
	m_nErrorVideos = 0;
	m_nErrorBraces = 0;

	// For unicode chars.
	CHAR	ansiBuf[3] = "\0";
	int		nAnsi = 0;

	// For non-string null value test.
	m_nCharsBetween = 0;
	CString strBetween;

	// Buffer.
	const UINT bufSz = 4096;
	char buf[bufSz];

	// Read chunks.
	while((nRead = datafile.Read(buf, bufSz)) > 0)
	{
		// Parse the chunk.
		for(UINT i = 0; i < nRead; i++)
		{
			ch = buf[i];
			iCh = buf[i];

			// Convert unicode.
			if(iCh < 1 || iCh > 255)
			{
				if(nAnsi < 2)
				{
					ansiBuf[nAnsi] = (char)ch;
					nAnsi ++;
					if(nAnsi == 1)
						continue;
				}
				if(nAnsi > 1)
				{
					TCHAR t;

					// Zero element - 2 chars max.
					nAnsi = 0;	
					
					// Get as wchar
					if(MultiByteToWideChar(CP_UTF8, 0, ansiBuf, 2, &t, 1))	// CP_UTF8 CP_ACP
					{
						ch = t;
					}

					// Can't use the chars.
					else
					{
						TRACE(_T("CKhanJson::ParseLibrary unicode convert failed\n"));
						continue;
					}
				}
			}

			// Switch the char.
			switch(iCh)
			{
				case 91:	// [
					nBrackets ++;
					break;

				case 93:	// ]
					nBrackets --;
					break;

				case 123:	// {
				{
					// Count braces.
					if(!inString)
						m_numBraces ++;

					// Count video block braces.
					if(m_inVideoBlock)
						m_nVidBraces ++;

					// Count Topic block braces.
					if(m_inTopicBlock)
						m_nTopicBraces ++;
				}
				break;

				case 125:	// }
				{
					// Count braces.
					if(!inString)
						m_numBraces --;

					// If inside a Video Block.
					m_nVidBraces --;
					if(m_inVideoBlock && !m_nVidBraces)
					{
						// Add a video to the current topic.
						// Note video may not be added if invalid.
						AddCurrentVideo();
						m_inVideoBlock = FALSE;
					}
					
					// If inside a Topic Block.
					m_nTopicBraces --;
					if(m_inTopicBlock && !m_nTopicBraces)
					{
						// Add the current topic to the tree.
						AddCurrentTopic(tree, hWndDlg, progPos);
						m_inTopicBlock = FALSE;
					}
				}
				break;				
				
				// Escape sequence next. v0.9.2.0
				case 92: // "\" 
				{
					// Save previous two.
					prevPrev_ch = prev_ch;
					prev_ch = ch;				
				}
				break;

				// Start or end of a string.
				case 34:	// "
				{
					// String is complete.
					if(inString)
					{
						// If an escaped single quote ", 
						if(prev_ch == _T('\\'))
						{
							// Get a ref.
							CString & str = getKeyword ? m_keyval.keyword : m_keyval.value;

							// If there is an escaped escape, add a "\"
							// and save. Example: "key": "\\". v0.9.2.0
							if(prevPrev_ch == _T('\\'))
							{
								str += _T('\\');
								inString = FALSE;
							}

							// Else, the " is part of a string.
							else
							{
								str += ch;

								// Get next char.
								prevPrev_ch = prev_ch;
								prev_ch = ch;
								break;
							}

							// Prep for init below.
							prev_ch = _T('0');
						}

						// If getting a keyword.
						if(getKeyword)
						{
							// If valid keyword, get a value next.
							if(IsKeyword())
							{
								getKeyword = FALSE;

								// Begin saving chars after a keyword.
								m_nCharsBetween = 0;
								strBetween = _T("");
							}
							// Else get another keyword.
							else
							{
								m_keyval.Zero();
							}
							// String is cessed.
							inString = FALSE;
						}

						// Getting a value.
						else
						{
							// Process the KEYVAL pair.
							ProcessKeyVal(tree);

							// String processed. Prepare for next.
							m_keyval.Zero();
							getKeyword = TRUE;
							inString = FALSE;
						}
					}

					// Else begin a string.
					else
					{
						// If begining to get a Key Value, check for a 
						// non-string null key value.
						// Note the distance between a keyWord and a keyValue
						// is usually 2 chars.
						if(!getKeyword && m_nCharsBetween > 4) // ": null,"
						{
							// If a null key value, we are getting a new keyWord.
							if( ( strBetween.Find(_T("null"), 0) ) != -1 )
							{
								getKeyword = TRUE;
								m_nCharsBetween = 0;
								strBetween = _T("");
							}
						}

						// Begining a string.
						inString = TRUE;
					}

					// Save previous chars.
					prevPrev_ch = prev_ch;
					prev_ch = ch;				
				}
				break;

				// Get chars.
				default:
				{
					// Get string data.
					if(inString)
					{
						// Escape sequences. See case 92 above. 11/03/13
						//
						// Add newlines.
						if(prev_ch == 92 && ch == 110)	// '\' and 'n'
						{
							if(getKeyword)
								m_keyval.keyword += prev_ch;
							else
								m_keyval.value += prev_ch;
						}
						// Add chars.
						if(getKeyword)
							m_keyval.keyword += ch;
						else
							m_keyval.value += ch;
					}
					// Get a video's "date_added" - a non-string value.
					else if(m_keyval.valueIsNumeric)
					{
						// Should end with a comma.
						if(ch == _T(','))
						{
							// Save the value.
							ProcessKeyVal(tree);

							// Prepare for next.
							m_keyval.Zero();
							getKeyword = TRUE;
						}
						// Save up the non-string chars.
						else
							m_keyval.value += ch;
					}
					// Gather up non string chars following a keyword.
					else if(!getKeyword)
					{
						m_nCharsBetween ++;
						strBetween += ch;
					}

					// Save char as previous
					prevPrev_ch = prev_ch;
					prev_ch = ch;					
				}
				break;				
			}//switch
		}
	}
	return TRUE;


/////////////////////////////////////
/*
	// File for reading.
	CFile datafile;
	CFileException e;
	if(!datafile.Open(pathIn, CFile::modeRead | CFile::typeBinary, &e))
	{
		TCHAR szError[1024];
		e.GetErrorMessage(szError, 1024);
		CString er;
		er.Format(_T("Couldn't open file: %s"), szError);
		TRACE(er);
		errMsg = er;
		return FALSE;
	}

	TCHAR	ch;
	TCHAR	prev_ch;
	int		iCh;
	UINT	nRead = 0;
	BOOL	inString = FALSE;
	BOOL	getKeyword = TRUE;
	int		nBrackets = 0;		// not used

	m_numBraces    = 0;
	m_inVideoBlock = FALSE;
	m_nVidBraces   = 1;
	m_inTopicBlock = FALSE;
	m_nTopicBraces = 1;

	// Error stats v0.9.0
	// Used in global KhanUpdateThread() in Doc.cpp
	// to possibly abort update.
	m_nErrorTopics = 0;
	m_nErrorVideos = 0;
	m_nErrorBraces = 0;

	// For unicode chars.
	CHAR	ansiBuf[3] = "\0";
	int		nAnsi = 0;

	// For non-string null value test.
	m_nCharsBetween = 0;
	CString strBetween;

	// Buffer.
	const UINT bufSz = 4096;
	char buf[bufSz];

	// Read chunks.
	while((nRead = datafile.Read(buf, bufSz)) > 0)
	{
		// Parse the chunk.
		for(UINT i = 0; i < nRead; i++)
		{
			ch = buf[i];
			iCh = buf[i];

			// Convert unicode.
			if(iCh < 1 || iCh > 255)
			{
				if(nAnsi < 2)
				{
					ansiBuf[nAnsi] = (char)ch;
					nAnsi ++;
					if(nAnsi == 1)
						continue;
				}
				if(nAnsi > 1)
				{
					TCHAR t;

					// Zero element - 2 chars max.
					nAnsi = 0;	
					
					// Get as wchar
					if(MultiByteToWideChar(CP_UTF8, 0, ansiBuf, 2, &t, 1))	// CP_UTF8 CP_ACP
						ch = t;

					// Can't use the chars.
					else
					{
					//	TRACE(_T("CKhanJson::ParseLibrary unicode convert failed\n"));
						continue;
					}
				}
			}

			// Switch the char.
			switch(iCh)
			{
				case 91:	// [
					nBrackets ++;
					break;

				case 93:	// ]
					nBrackets --;
					break;

				case 123:	// {
				{
					// Count braces.
					if(!inString)
						m_numBraces ++;

					// Count video block braces.
					if(m_inVideoBlock)
						m_nVidBraces ++;

					// Count Topic block braces.
					if(m_inTopicBlock)
						m_nTopicBraces ++;
				}
				break;

				case 125:	// }
				{
					// Count braces.
					if(!inString)
						m_numBraces --;

					// If inside a Video Block.
					m_nVidBraces --;
					if(m_inVideoBlock && !m_nVidBraces)
					{
						// Add a video to the current topic.
						// Note video may not be added if invalid.
						AddCurrentVideo();
						m_inVideoBlock = FALSE;
					}
					
					// If inside a Topic Block.
					m_nTopicBraces --;
					if(m_inTopicBlock && !m_nTopicBraces)
					{
						// Add the current topic to the tree.
						AddCurrentTopic(tree, hWndDlg, progPos);
						m_inTopicBlock = FALSE;
					}
				}
				break;

				// Start or end of a string.
				case 34:	// "
				{
					// Process string data.
					if(inString)
					{
						// If getting a keyword.
						if(getKeyword)
						{
							// If an escaped ". Probably only in keyval's below.
							if(prev_ch == _T('\\'))
							{
								prev_ch = ch;
								m_keyval.keyword += ch;
								break;
							}

							// If valid keyword, get a value next.
							else if(IsKeyword())
							{
								getKeyword = FALSE;

								// Begin saving chars after a keyword.
								m_nCharsBetween = 0;
								strBetween = _T("");
							}
							// Else get another keyword.
							else
							{
								m_keyval.Zero();
							}
						}
						// Getting a value
						else
						{
							// If an escaped ".
							if(prev_ch == _T('\\'))
							{
								prev_ch = ch;
								m_keyval.value += ch;
								break;
							}
							// Process the KEYVAL pair.
							ProcessKeyVal(tree);

							// Prepare for next.
							m_keyval.Zero();
							getKeyword = TRUE;
						}
						// String processed.
						inString = FALSE;
					}

					// Get string data.
					else
					{
						// If begining to get a Key Value, check for a 
						// non-string null key value.
						// Note the distance between a keyWord and a keyValue
						// is usually 2 chars.
						if(!getKeyword && m_nCharsBetween > 4) // ": null,"
						{
							// If a null key value, we are getting a new keyWord.
							if( ( strBetween.Find(_T("null"), 0) ) != -1 )
							{
								getKeyword = TRUE;
								m_nCharsBetween = 0;
								strBetween = _T("");
							}
						}

						// Begining a string.
						inString = TRUE;
					}					
				}
				break;

				// Get chars.
				default:
				{
					// Get string data.
					if(inString)
					{
						if(getKeyword)
							m_keyval.keyword += ch;
						else
							m_keyval.value += ch;
					}
					// Get a video's "date_added" - a non-string value.
					else if(m_keyval.valueIsNumeric)
					{
						// Should end with a comma.
						if(ch == _T(','))
						{
							// Save the value.
							ProcessKeyVal(tree);

							// Prepare for next.
							m_keyval.Zero();
							getKeyword = TRUE;
						}
						// Save up the non-string chars.
						else
							m_keyval.value += ch;
					}
					// Gather up non string chars following a keyword.
					else if(!getKeyword)
					{
						m_nCharsBetween ++;
						strBetween += ch;
					}

					// Save char as previous
					prev_ch = ch;					
				}
				break;				
			}//switch
		}
	}
	return TRUE;*/
}

// Check if m_keyval.keyword a Khan Keyword.
BOOL CKhanJson::IsKeyword()
{
	// date_added
	if(m_keyval.keyword == _T("date_added"))
		return TRUE;

	// description
	else if(m_keyval.keyword == _T("description"))
		return TRUE;

	// mp4
	else if(m_keyval.keyword == _T("mp4"))
		return TRUE;

	// duration
	else if(m_keyval.keyword == _T("duration"))
	{
		m_keyval.valueIsNumeric = TRUE;
		return TRUE;
	}

	// title - Video 
	else if(m_keyval.keyword == _T("title"))
		return TRUE;

	// kind
	else if(m_keyval.keyword == _T("kind"))
		return TRUE;

	// standalone_title - topic
	else if(m_keyval.keyword == _T("standalone_title"))
		return TRUE;

	// id
	else if(m_keyval.keyword == _T("id"))
		return TRUE;

	// slug 
	else if(m_keyval.keyword == _T("slug"))
		return TRUE;

	// extended_slug 
	else if(m_keyval.keyword == _T("extended_slug"))
		return TRUE;

	return FALSE;
}

// Process a KEYVAL pair.
BOOL CKhanJson::ProcessKeyVal(CTopicTree * tree)
{
	// date_added
	if(m_keyval.keyword == _T("date_added"))
		m_dateAdded = m_keyval.value;

	// description - topic or video.
	else if(m_keyval.keyword == _T("description"))
		m_description = m_keyval.value;

	// mp4
	else if(m_keyval.keyword == _T("mp4"))
		m_mp4 = m_keyval.value;

	// duration
	else if(m_keyval.keyword == _T("duration"))
	{
		m_duration = m_keyval.value;
		// Clean up string, e.g., ": 339"
		m_duration.Remove(_T(':'));
		m_duration.TrimLeft();
		m_duration.TrimRight();
	}

	// title - Video 
	else if(m_keyval.keyword == _T("title"))
	{
		m_title = m_keyval.value;
		// Ensure no spaces at end. Fix v0.8.5 - for a few vid postfix errors
		m_title.TrimRight();

		// Check for the root topic title. v0.9.1
		// See also id test below.
		if(m_title == _T("The Root of All Knowledge"))
		{
			m_numBraces = 0;
			ZeroCurrentTopic();
			m_title = _T("");
			return FALSE;
		}
	}

	// kind
	else if(m_keyval.keyword == _T("kind"))
	{
		// "Video"
		if(m_keyval.value == _T("Video"))
		{
			// Getting a video.
			m_inVideoBlock = TRUE;
			m_nVidBraces = 1;
		}
		// "Topic"
		else if(m_keyval.value == _T("Topic"))
		{
			// Getting a Topic.
			m_inTopicBlock = TRUE;
			m_nTopicBraces = 1;
		}
	}

	// "id"
	else if(m_keyval.keyword == _T("id"))
	{
		// Not saving the root.
		// See also title test above.
		if(m_keyval.value == _T("x00000000")) // v0.9.1
		{
			m_numBraces = 0;
			TestZeroCurrentTopic();
			return FALSE;
		}
	}

	// standalone_title - topic
	else if(m_keyval.keyword == _T("standalone_title"))
	{
		m_standaloneTitle = m_keyval.value;
	}

	// extended_slug - topic - v0.9.1
	else if(m_keyval.keyword == _T("extended_slug"))
		 m_extendedSlug = m_keyval.value;

	return TRUE;
}

// Add the m_curTopic member to the tree.
BOOL CKhanJson::AddCurrentTopic(CTopicTree * tree, HWND hWndDlg, int & progPos)
{
	// If at the end.
	if(m_numBraces < 0)
	{
		TRACE(_T("WARNING CKhanJson::AddCurrentTopic - m_numBraces < 0.\n"));
		ZeroCurrentTopic();
		return FALSE;
	}

	// Get nodes list.
	if(!SetTopicNodes())
	{
		ZeroCurrentTopic();
		return FALSE;
	}

	// Get num topic nodes.
	int nNodes = m_nodesAr.GetSize();
   
	// Set the topic's title and description.
	m_curTopic.title = m_standaloneTitle;
	m_curTopic.description = m_description;

	// Set this topics tree topic to the last id array element.
	m_curTopic.treeTopic = m_nodesAr[nNodes - 1];

	// Set this topics parent tree topic.
	if(nNodes > 1)
		m_curTopic.parentTreeTopic = m_nodesAr[nNodes - 2];

	// Add the topic...
	//
	// Set some flags.
	BOOL bVids	= m_curTopic.videos.GetSize() > 0;
	BOOL bTitle = !m_curTopic.title.IsEmpty();
	BOOL bDescription = !m_curTopic.description.IsEmpty();

	// If a topic with no videos.
	if(!bVids)
	{
		// We add topics to the tree only if they have videos.
		// Parents of the topics, named in the node list, have
		// no videos and are added as emptys. 
		//
		// For example "math" in math/algebra.
		//
		// If we have a Title str and or a Definition str,
		// use them to updtate the parent topic.
		// 
		if(bTitle || bDescription)
		{
			tree->UpdateTopic(m_curTopic, m_nodesAr, TRUE);

			// Prepare for next.
			ZeroCurrentTopic();

			return TRUE;
		}
		// Else an empty topic - discard it.
		//
		TRACE(_T("ERROR CKhanJson::AddCurrentTopic - topic invalid.\n"));

		// Prepare for next.
		ZeroCurrentTopic();

		return FALSE;
	}

	// Videos may not have their topic string set.
	int nVids = m_curTopic.videos.GetSize();
	for(int i = 0; i < nVids; i++)
	{
		m_curTopic.videos[i].topic = m_curTopic.treeTopic;
	}

	// Add to Topic Tree Param.
	tree->AddTopic(m_curTopic, m_nodesAr);

	// Prepare for next.
	ZeroCurrentTopic();

	// Update dlg progress pos every 20 topics.
	// Note 500 + topics as of Jan 2013.
	if(hWndDlg)
	{
		progPos ++;
		if(progPos % 20 == 0)
			PostMessage(hWndDlg, WM_USER_PROGRESS_POS, 0, (LPARAM)progPos);
	}

	return TRUE;
}

// Set the topic nodes per extended_slug. v0.9.1
BOOL CKhanJson::SetTopicNodes()
{
	// Prepare.
	m_nodesAr.RemoveAll();

	// Must have.
	if(m_extendedSlug.IsEmpty())
		return FALSE;

	CString str = m_extendedSlug;

	// Ensure lower case.
	str.MakeLower();

	// Parse slug: e.g "math/algebra/absolute-value"
	CString sub;
	int i;
	while(( i = str.Find(_T("/"), 0) ) != -1)
	{
		sub = str.Left(i);
		str.Delete(0, i + 1);
		m_nodesAr.Add(sub);
	}
	// Add the last
	m_nodesAr.Add(str);

	// Validate.
	int nNodes = m_nodesAr.GetSize();

	// Should always have one node greater
	// than current brace count.
	// This will abort the update - see Doc.cpp KhanUpdateThread().
	if(m_numBraces + 1 != nNodes)
	{
		TRACE(_T("ERROR Topic slugs do not match brace count\n"));
		m_nodesAr.RemoveAll();
		m_nErrorBraces ++;
		return FALSE;
	}

	// Ok.
	return TRUE;
}


// Add the m_curVideo member to the Current Topic (m_curTopic).
BOOL CKhanJson::AddCurrentVideo()
{
	// Set video data.
	m_curVideo.date_added = m_dateAdded;
	m_curVideo.description = m_description;
	m_curVideo.mp4 = m_mp4;
	m_curVideo.duration = m_duration;
	m_curVideo.title = m_title;

	// Some videos now come with no video! 34 videos on 10/03/12.
	// This is indicated by - "download_urls": null
	if(m_curVideo.mp4.IsEmpty())
	{
		// Prepare for next and bail.
		m_curVideo.Zero();
		return FALSE;		// return val not important		
	}

	// Add to topic array.
	m_curTopic.videos.Add(m_curVideo);

	// Prepare for next.
	m_curVideo.Zero();
	m_dateAdded = _T("");
	m_description = _T("");
	m_mp4 = _T("");
	m_duration = _T("");
	m_title = _T("");

	return TRUE;
}

// Prepare Current Topic data.
void CKhanJson::ZeroCurrentTopic()
{
	m_curTopic.Zero();
	m_standaloneTitle = _T("");
	m_description = _T("");
	m_extendedSlug = _T("");
}


//////////////////////////////////////////////////////////////////////////
// Test for /api/v1/topictree

/* Profile: release, 673 topics, 10536 kb json file, 03/19/13

	   763.054 milliseconds
       100.0 %        1 CKhanJson::TestParseLibrary
        76.3	6108185 CString::operator+=
         6.1     534942 CString::operator=
         4.4     250506 CKhanJson::TestIsKeyword
*/

BOOL CKhanJson::TestParseLibrary(CTopicTree * tree, const CString & pathIn)
{
	// Params used by non test version.
	HWND hWndDlg = NULL;
	int progPos = 0;
	CString errMsg;

	// File for reading.
	CFile datafile;
	CFileException e;
	if(!datafile.Open(pathIn, CFile::modeRead | CFile::typeBinary, &e))
	{
		TCHAR szError[1024];
		e.GetErrorMessage(szError, 1024);
		CString er;
		er.Format(_T("Couldn't open file: %s"), szError);
		TRACE(er);
		errMsg = er;
		return FALSE;
	}

	TCHAR	ch = _T('0');
	TCHAR	prev_ch = _T('0');
	TCHAR	prevPrev_ch = _T('0');
	int		iCh = 0;
	UINT	nRead = 0;
	BOOL	inString = FALSE;
	BOOL	getKeyword = TRUE;
	int		nBrackets = 0;		// not used

	m_numBraces    = 0;
	m_inVideoBlock = FALSE;
	m_nVidBraces   = 1;
	m_inTopicBlock = FALSE;
	m_nTopicBraces = 1;

	// Error stats v0.9.0
	// Used in global KhanUpdateThread() in Doc.cpp
	// to possibly abort update.
	m_nErrorTopics = 0;
	m_nErrorVideos = 0;
	m_nErrorBraces = 0;

	// For unicode chars.
	CHAR	ansiBuf[3] = "\0";
	int		nAnsi = 0;

	// For non-string null value test.
	m_nCharsBetween = 0;
	CString strBetween;

	// Buffer.
	const UINT bufSz = 4096;
	char buf[bufSz];

	// Read chunks.
	while((nRead = datafile.Read(buf, bufSz)) > 0)
	{
		// Parse the chunk.
		for(UINT i = 0; i < nRead; i++)
		{
			ch = buf[i];
			iCh = buf[i];

			// Convert unicode.
			if(iCh < 1 || iCh > 255)
			{
				if(nAnsi < 2)
				{
					ansiBuf[nAnsi] = (char)ch;
					nAnsi ++;
					if(nAnsi == 1)
						continue;
				}
				if(nAnsi > 1)
				{
					TCHAR t;

					// Zero element - 2 chars max.
					nAnsi = 0;	
					
					// Get as wchar
					if(MultiByteToWideChar(CP_UTF8, 0, ansiBuf, 2, &t, 1))	// CP_UTF8 CP_ACP
					{
						ch = t;
					}

					// Can't use the chars.
					else
					{
						TRACE(_T("CKhanJson::TestParseLibrary unicode convert failed\n"));
						continue;
					}
				}
			}

			// Switch the char.
			switch(iCh)
			{
				case 91:	// [
					nBrackets ++;
					break;

				case 93:	// ]
					nBrackets --;
					break;

				case 123:	// {
				{
					// Count braces.
					if(!inString)
						m_numBraces ++;

					// Count video block braces.
					if(m_inVideoBlock)
						m_nVidBraces ++;

					// Count Topic block braces.
					if(m_inTopicBlock)
						m_nTopicBraces ++;
				}
				break;

				case 125:	// }
				{
					// Count braces.
					if(!inString)
						m_numBraces --;

					// If inside a Video Block.
					m_nVidBraces --;
					if(m_inVideoBlock && !m_nVidBraces)
					{
						// Add a video to the current topic.
						// Note video may not be added if invalid.
						TestAddCurrentVideo();
						m_inVideoBlock = FALSE;
					}
					
					// If inside a Topic Block.
					m_nTopicBraces --;
					if(m_inTopicBlock && !m_nTopicBraces)
					{
						// Add the current topic to the tree.
						TestAddCurrentTopic(tree, hWndDlg, progPos);
						m_inTopicBlock = FALSE;
					}
				}
				break;				
				
				// Escape sequence next. v0.9.2.0
				case 92: // "\" 
				{
					// Save previous two.
					prevPrev_ch = prev_ch;
					prev_ch = ch;				
				}
				break;

				// Start or end of a string.
				case 34:	// "
				{
					// bug: Preventing TB using the \"4 I's\"
					// new
					// String is complete.
					if(inString)
					{
						// If an escaped single quote ", 
						if(prev_ch == _T('\\'))
						{
							// Get a ref.
							CString & str = getKeyword ? m_keyval.keyword : m_keyval.value;

							// If there is an escaped escape, add a "\"
							// and save. Example: "key": "\\". v0.9.2.0
							if(prevPrev_ch == _T('\\'))
							{
								str += _T('\\');
								inString = FALSE;
							}
							// Else, the " is part of a string.
							else
							{
								str += ch;

								// Get next char.
								prevPrev_ch = prev_ch;
								prev_ch = ch;
								break;
							}
							// Prep for init below.
							prev_ch = _T('0');
						}

						// If getting a keyword.
						if(getKeyword)
						{
							// If valid keyword, get a value next.
							if(TestIsKeyword())
							{
								getKeyword = FALSE;

								// Begin saving chars after a keyword.
								m_nCharsBetween = 0;
								strBetween = _T("");
							}
							// Else get another keyword.
							else
							{
								m_keyval.Zero();
							}
							// String is cessed.
							inString = FALSE;
						}

						// Getting a value.
						else
						{
							// Process the KEYVAL pair.
							TestProcessKeyVal(tree);

							// String processed. Prepare for next.
							m_keyval.Zero();
							getKeyword = TRUE;
							inString = FALSE;
						}
					}

					// Else begin a string.
					else
					{
						// If begining to get a Key Value, check for a 
						// non-string null key value.
						// Note the distance between a keyWord and a keyValue
						// is usually 2 chars.
						if(!getKeyword && m_nCharsBetween > 4) // ": null,"
						{
							// If a null key value, we are getting a new keyWord.
							if( ( strBetween.Find(_T("null"), 0) ) != -1 )
							{
								getKeyword = TRUE;
								m_nCharsBetween = 0;
								strBetween = _T("");
							}
						}

						// Begining a string.
						inString = TRUE;
					}

					// Save previous chars.
					prevPrev_ch = prev_ch;
					prev_ch = ch;				
				}
				break;

				// Get chars.
				default:
				{
					// Get string data.
					if(inString)
					{
						// Escape sequences. See case 92 above. 11/03/13
						//
						// Add newlines.
						if(prev_ch == 92 && ch == 110)	// '\' and 'n'
						{
							if(getKeyword)
								m_keyval.keyword += prev_ch;
							else
								m_keyval.value += prev_ch;
						}
						// Add chars.
						if(getKeyword)
							m_keyval.keyword += ch;
						else
							m_keyval.value += ch;
					}
					// Get a video's "date_added" - a non-string value.
					else if(m_keyval.valueIsNumeric)
					{
						// Should end with a comma.
						if(ch == _T(','))
						{
							// Save the value.
							TestProcessKeyVal(tree);

							// Prepare for next.
							m_keyval.Zero();
							getKeyword = TRUE;
						}
						// Save up the non-string chars.
						else
							m_keyval.value += ch;
					}
					// Gather up non string chars following a keyword.
					else if(!getKeyword)
					{
						m_nCharsBetween ++;
						strBetween += ch;
					}

					// Save char as previous
					prevPrev_ch = prev_ch;
					prev_ch = ch;					
				}
				break;				
			}//switch
		}
	}
	return TRUE;
}

BOOL CKhanJson::TestIsKeyword()
{
	// date_added
	if(m_keyval.keyword == _T("date_added"))
		return TRUE;

	// description
	else if(m_keyval.keyword == _T("description"))
		return TRUE;

	// mp4
	else if(m_keyval.keyword == _T("mp4")) // "download_urls" ?
		return TRUE;

	// duration
	else if(m_keyval.keyword == _T("duration"))
	{
		m_keyval.valueIsNumeric = TRUE;
		return TRUE;
	}

	// title - Video 
	else if(m_keyval.keyword == _T("title"))
		return TRUE;

	// kind
	else if(m_keyval.keyword == _T("kind"))
		return TRUE;

	// standalone_title - topic
	else if(m_keyval.keyword == _T("standalone_title"))
		return TRUE;

	// id
	else if(m_keyval.keyword == _T("id"))
		return TRUE;

	// slug 
	else if(m_keyval.keyword == _T("slug"))
		return TRUE;

	// extended_slug 
	else if(m_keyval.keyword == _T("extended_slug"))
		return TRUE;

	return FALSE;
}

// new
BOOL CKhanJson::TestProcessKeyVal(CTopicTree * tree)
{
	// date_added
	if(m_keyval.keyword == _T("date_added"))
		m_dateAdded = m_keyval.value;

	// description - topic or video.
	else if(m_keyval.keyword == _T("description"))
		m_description = m_keyval.value;

	// mp4
	else if(m_keyval.keyword == _T("mp4"))
		m_mp4 = m_keyval.value;

	// duration
	else if(m_keyval.keyword == _T("duration"))
	{
		m_duration = m_keyval.value;
		// Clean up string, e.g., ": 339"
		m_duration.Remove(_T(':'));
		m_duration.TrimLeft();
		m_duration.TrimRight();
	}

	// title - Video 
	else if(m_keyval.keyword == _T("title"))
	{
		m_title = m_keyval.value;
		// Ensure no spaces at end. Fix v0.8.5 - for a few vid postfix errors
		m_title.TrimRight();

		// Check for the root topic title. v0.9.1
		// See also id test below.
		if(m_title == _T("The Root of All Knowledge"))
		{
			m_numBraces = 0;
			TestZeroCurrentTopic();
			m_title = _T("");
			return FALSE;
		}
	}

	// kind
	else if(m_keyval.keyword == _T("kind"))
	{
		// "Video"
		if(m_keyval.value == _T("Video"))
		{
			// Getting a video.
			m_inVideoBlock = TRUE;
			m_nVidBraces = 1;
		}
		// "Topic"
		else if(m_keyval.value == _T("Topic"))
		{
			// Getting a Topic.
			m_inTopicBlock = TRUE;
			m_nTopicBraces = 1;
		}
	}

	// "id"
	// bug - v0.9.0 - Now we get null id's - 30 or so
	//
	// v0.9.1 - id's are all numeric now - back to using 
	// extended_slug instead - api hell.
	else if(m_keyval.keyword == _T("id"))
	{
		// Not saving the root.
		// See also title test above.
		if(m_keyval.value == _T("x00000000")) // v0.9.1
		{
			m_numBraces = 0;
			TestZeroCurrentTopic();
			return FALSE;
		}
	}

	// standalone_title - topic
	else if(m_keyval.keyword == _T("standalone_title"))
	{
		m_standaloneTitle = m_keyval.value;
	}

	// extended_slug - topic - v0.9.0
	else if(m_keyval.keyword == _T("extended_slug"))
	{
		 m_extendedSlug = m_keyval.value;
	}

	return TRUE;
}

BOOL CKhanJson::TestAddCurrentTopic(CTopicTree * tree, HWND hWndDlg, int & progPos)
{
	// If at the end.
	if(m_numBraces < 0)
	{
		TRACE(_T("WARNING CKhanJson::TestAddCurrentTopic - m_numBraces < 0.\n"));
		TestZeroCurrentTopic();
		return FALSE;
	}

	// Get nodes list.
	if(!TestSetTopicNodes())
	{
		TestZeroCurrentTopic();
		return FALSE;
	}

	// Get num topic nodes.
	int nNodes = m_nodesAr.GetSize();

	// dbg
	CString s; for(int n = 0; n < nNodes; n ++) { s += m_nodesAr[n]; s += _T('\n'); }
   
	// Set the topic's title and description.
	m_curTopic.title = m_standaloneTitle;
	m_curTopic.description = m_description;

	// dbg - save raw slugs - See TestParseLibrary() and CDataView::OnTest().
	CString s2; for(int n2 = 0; n2 < nNodes; n2 ++) { s2 += m_nodesAr[n2]; s2 += _T('/'); }
	m_sortStrAr.Add(s2);

	// Set this topics tree topic to the last id array element.
	m_curTopic.treeTopic = m_nodesAr[nNodes - 1];

	// Set this topics parent tree topic.
	if(nNodes > 1)
		m_curTopic.parentTreeTopic = m_nodesAr[nNodes - 2];

	// Add the topic...
	//
	// Set some flags.
	BOOL bVids	= m_curTopic.videos.GetSize() > 0;
	BOOL bTitle = !m_curTopic.title.IsEmpty();
	BOOL bDescription = !m_curTopic.description.IsEmpty();

	// If a topic with no videos.
	if(!bVids)
	{
		// We add topics to the tree only if they have videos.
		// Parents of the topics, named in the node list, have
		// no videos and are added as emptys. 
		//
		// For example "math" in math/algebra.
		//
		// If we have a Title str and or a Definition str,
		// use them to updtate the parent topic.
		// 
		if(bTitle || bDescription)
		{
			tree->UpdateTopic(m_curTopic, m_nodesAr, TRUE);

			// Prepare for next.
			TestZeroCurrentTopic();

			return TRUE;
		}
		// Else an empty topic - discard it.
		//
		TRACE(_T("ERROR CKhanJson::TestAddCurrentTopic - topic invalid.\n"));

		// Prepare for next.
		TestZeroCurrentTopic();

		return FALSE;
	}

	// Videos may not have their topic string set.
	int nVids = m_curTopic.videos.GetSize();
	for(int i = 0; i < nVids; i++)
	{
		m_curTopic.videos[i].topic = m_curTopic.treeTopic;
	}

	// Add to Topic Tree Param.
	tree->AddTopic(m_curTopic, m_nodesAr);

	// Prepare for next.
	TestZeroCurrentTopic();

	// Update dlg progress pos every 20 topics.
	// Note 500 + topics as of Jan 2013.
	if(hWndDlg)
	{
		progPos ++;
		if(progPos % 20 == 0)
			PostMessage(hWndDlg, WM_USER_PROGRESS_POS, 0, (LPARAM)progPos);
	}

	return TRUE;
}


// Set the topic nodes per extended_slug. v0.9.1
BOOL CKhanJson::TestSetTopicNodes()
{
	// Prepare.
	m_nodesAr.RemoveAll();

	// Must have.
	if(m_extendedSlug.IsEmpty())
		return FALSE;

	CString str = m_extendedSlug;

	// Ensure lower case.
	str.MakeLower();

	// Parse slug: e.g "math/algebra/absolute-value"
	CString sub;
	int i;
	while(( i = str.Find(_T("/"), 0) ) != -1)
	{
		sub = str.Left(i);
		str.Delete(0, i + 1);
		m_nodesAr.Add(sub);
	}
	// Add the last
	m_nodesAr.Add(str);

	// Validate.
	int nNodes = m_nodesAr.GetSize();

	// dbg - save
	CString s; for(int n = 0; n < nNodes; n ++) { s += m_nodesAr[n]; s += _T('\n'); }

	// Should always have one node greater
	// than current brace count.
	// This will abort the update - see Doc.cpp KhanUpdateThread().
	if(m_numBraces + 1 != nNodes)
	{
		CString msg;
		msg.Format(_T("Brace error at exSlug: %s\n"), m_extendedSlug);
		TRACE(msg);
		m_nodesAr.RemoveAll();
		m_nErrorBraces ++;
		return FALSE;
	}

	// Ok.
	return TRUE;
}

// Prepare Current Topic data.
void CKhanJson::TestZeroCurrentTopic()
{
	m_curTopic.Zero();
	m_standaloneTitle = _T("");
	m_description = _T("");
	m_extendedSlug = _T("");
}

BOOL CKhanJson::TestAddCurrentVideo()
{
	// Set video data.
	m_curVideo.date_added = m_dateAdded;
	m_curVideo.description = m_description;
	m_curVideo.mp4 = m_mp4;
	m_curVideo.duration = m_duration;
	m_curVideo.title = m_title;


	// Some videos now come with no video! 
	//
	//	34 videos on 10/03/12.
	//	2386 videos on 11/03/13. wtf?
	//
	// This is indicated by - "download_urls": null, or "download_urls": {}
	if(m_curVideo.mp4.IsEmpty())
	{
		// Write title to debug output:
		CString str;
		str.Format(_T("No mp4 for: %s\n"), m_title);
		TRACE(str);

		// Prepare for next and bail.
		m_curVideo.Zero();
		m_nErrorVideos ++;
		return FALSE;		// return val not important		
	}

	// Add to topic array.
	m_curTopic.videos.Add(m_curVideo);

	// Prepare for next.
	m_curVideo.Zero();
	m_dateAdded = _T("");
	m_description = _T("");
	m_mp4 = _T("");
	m_duration = _T("");
	m_title = _T("");

	return TRUE;
}


/////////////////////////////////////////////////////////////////////////

// Gross format json data to make it more readable 
// for debugging parse routines.
//
// pathIn is binary, pathOut txt.
void CKhanJson::TestFormat(CString pathIn, CString pathOut)
{
	CString lineStr1, lineStr2, lineStr3;
	
	// File for reading.
	CFile srcfile;
	CFileException e;
	if(!srcfile.Open(pathIn, CFile::modeRead | CFile::typeBinary, &e))
	{
		TCHAR szError[1024];
		e.GetErrorMessage(szError, 1024);
		CString er;
		er.Format(_T("Couldn't open file for reading: %s"), szError);
		TRACE(er);
		return;
	}

	// File for writing.
	CStdioFile destfile;
	if(!destfile.Open(pathOut, CFile::modeWrite | CFile::modeCreate | 
							   CFile::typeText, &e))
	{
		TCHAR szError[1024];
		e.GetErrorMessage(szError, 1024);
		CString er;
		er.Format(_T("Couldn't open file for writing: %s"), szError);
		TRACE(er);
		return;
	}

	const int bufSz = 4096;
	char buf[bufSz];
	TCHAR	c;
	int		nRead;
	CString tab = _T('\t');
	CString crlf = _T('\n');
	int nTabs = 0;
	int t;
	BOOL inString = FALSE;
	int nBrackets = 0;
	CString strNumBracket;

	// Read chunks.
	while((nRead = srcfile.Read(buf, bufSz)) > 0)
	{
		// Format the chunk.
		for(int i = 0; i < nRead; i++) // 0x0D, 0x0A
		{
			c = buf[i];

			// Inside a string.
			if(c == _T('"'))
				inString = !inString;

			if(c == _T('{') && !inString)
			{
				nBrackets++;
				destfile.WriteString(crlf);

				for(t = 0; t < nTabs; t++)
					destfile.WriteString(tab);

				destfile.Write(&c, 1);

				strNumBracket.Format(_T("-%d"), nBrackets);
				destfile.WriteString(strNumBracket);

				destfile.WriteString(crlf);

				nTabs ++;

				for(t = 0; t < nTabs; t++)
					destfile.WriteString(tab);
			}
			else if(c == _T('}') && !inString)
			{
				nBrackets--;
				nTabs --;

				destfile.WriteString(crlf);

				for(t = 0; t < nTabs; t++)
					destfile.WriteString(tab);

				destfile.Write(&c, 1);

				strNumBracket.Format(_T("-%d"), nBrackets);
				destfile.WriteString(strNumBracket);

				destfile.WriteString(crlf);
			}
			else if(c == _T(',') && !inString)
			{
				destfile.Write(&c, 1);
				destfile.WriteString(crlf);
				for(t = 0; t < nTabs; t++)
					destfile.WriteString(tab);
			}

			else
				destfile.Write(&c, 1);
		}
	}
	// Clean up.
	srcfile.Close();
	destfile.Close();
}

// Print slugs	
void CKhanJson::TestPrintSlugs(CString pathOut, BOOL sort)
{	
	// File for writing.
	CStdioFile destfile;
	CFileException e;
	if(!destfile.Open(pathOut, CFile::modeWrite | CFile::modeCreate | 
							   CFile::typeText, &e))
	{
		TCHAR szError[1024];
		e.GetErrorMessage(szError, 1024);
		CString er;
		er.Format(_T("Couldn't open file for writing slugs: %s"), szError);
		TRACE(er);
		return;
	}
	CString crlf = _T('\n');

	// Sort.
	if(sort)
		m_sortStrAr.SortStrDecend();

	// Loop array.
	int sz = m_sortStrAr.GetSize();
	for(int i = 0 ; i < sz; i ++)
	{
		destfile.WriteString(m_sortStrAr[i]);
		destfile.WriteString(crlf);
	}

	// Clean up.
	destfile.Close();
}

